# -*- coding: utf-8 -*-
import os

from kiara.models.filesystem import KiaraFileBundle
from kiara.models.values.value import Value


def check_downloaded_file(file_bundle: Value):

    assert (
        file_bundle.data_type_name == "file_bundle"
    ), f"Expected a file_bundle value, got: {file_bundle.data_type_name}"
    assert (
        file_bundle.value_size > 144000
    ), f"Expected a file_bundle value with size > 144000, got: {file_bundle.value_size}"

    kiara_file_bundle: KiaraFileBundle = file_bundle.data
    assert (
        kiara_file_bundle.__class__ == KiaraFileBundle
    ), f"Expected a KiaraFileBundle object, got: {kiara_file_bundle.__class__}"

    assert (
        f"kiara_plugin{os.path.sep}core_types{os.path.sep}defaults.py"
        in kiara_file_bundle.included_files.keys()
    ), f"Expected 'kiara_plugin{os.path.sep}core_types{os.path.sep}defaults.py' in included files, got: {kiara_file_bundle.included_files.keys()}"



# -*- coding: utf-8 -*-
from kiara.models.filesystem import KiaraFile
from kiara.models.values.value import Value


def check_downloaded_file(file: Value):

    assert file.data_type_name == "file"
    assert file.value_size > 2000

    kiara_file: KiaraFile = file.data
    assert kiara_file.__class__ == KiaraFile

    assert "# kiara" in kiara_file.read_text()



# -*- coding: utf-8 -*-
import abc
import atexit
import shutil
import tempfile
from typing import TYPE_CHECKING, Any, Dict, List, Union

from pydantic import Field

from kiara.exceptions import KiaraException, KiaraProcessingException
from kiara.models.module import KiaraModuleConfig
from kiara.models.values.value import ValueMap
from kiara.modules import KiaraModule, ValueMapSchema

if TYPE_CHECKING:
    from kiara.models.filesystem import FolderImportConfig, KiaraFile, KiaraFileBundle


class OnboardFileConfig(KiaraModuleConfig):

    result_file_name: Union[str, None] = Field(
        description="The file name to use for the downloaded file, if not provided it will be auto-generated.",
        default=None,
    )
    attach_metadata: Union[bool, None] = Field(
        description="Whether to attach metadata. If 'None', a user input will be created.",
        default=True,
    )


ONBOARDING_MODEL_NAME_PREFIX = "onboarding.file.from."


class OnboardFileModule(KiaraModule):
    """A generic module that imports a file from one of several possible sources."""

    _config_cls = OnboardFileConfig

    @abc.abstractmethod
    def create_onboard_inputs_schema(self) -> Dict[str, Any]:
        pass

    def create_inputs_schema(
        self,
    ) -> ValueMapSchema:

        result = self.create_onboard_inputs_schema()

        if "file_name" in result.keys():
            raise KiaraException(
                msg="The 'file_name' input is not allowed in the onboard inputs schema for an implementing class of 'OnboardFileModule'."
            )
        result["file_name"] = {
            "type": "string",
            "doc": "The file name to use for the downloaded file, if not provided it will be auto-generated.",
            "optional": True,
        }

        if "attach_metadata" in result.keys():
            raise KiaraException(
                msg="The 'attach_metadata' input is not allowed in the onboard inputs schema for an implementing class of 'OnboardFileModule'."
            )

        file_name = self.get_config_value("result_file_name")
        if file_name is None:
            result["file_name"] = {
                "type": "string",
                "doc": "The file name to use for the downloaded file, if not provided it will be generated from the last token of the url.",
                "optional": True,
            }

        attach_metadata = self.get_config_value("attach_metadata")
        if attach_metadata is None:
            result["attach_metadata"] = {
                "type": "boolean",
                "doc": "Whether to attach onboarding metadata to the result file.",
                "default": True,
            }

        return result

    def create_outputs_schema(
        self,
    ) -> ValueMapSchema:

        result = {"file": {"type": "file", "doc": "The file that was onboarded."}}

        return result

    @abc.abstractmethod
    def retrieve_file(
        self, inputs: ValueMap, file_name: Union[str, None], attach_metadata: bool
    ) -> "KiaraFile":
        pass

    def process(self, inputs: ValueMap, outputs: ValueMap):

        if "file_name" not in inputs.keys():
            file_name = self.get_config_value("file_name")
        else:
            file_name = inputs.get_value_data("file_name")

        if "attach_metadata" not in inputs.keys():
            # must be 'True' or 'False', otherwise we'd have an input
            attach_metadata: bool = self.get_config_value("attach_metadata")
        else:
            attach_metadata = inputs.get_value_data("attach_metadata")

        result = self.retrieve_file(inputs, file_name, attach_metadata)

        outputs.set_value("file", result)


class OnboardFileBundleConfig(KiaraModuleConfig):

    result_bundle_name: Union[str, None] = Field(
        description="The bundle name use for the downloaded file, if not provided it will be auto-generated.",
        default=None,
    )

    attach_metadata_to_bundle: Union[bool, None] = Field(
        description="Whether to attach the download metadata to the result file bundle instance. If 'None', a user input will be created.",
        default=True,
    )
    attach_metadata_to_files: Union[bool, None] = Field(
        description="Whether to attach the download metadata to each file in the resulting bundle.",
        default=False,
    )
    sub_path: Union[None, str] = Field(description="The sub path to use.", default=None)
    include_files: Union[None, List[str]] = Field(
        description="List of file types (strings) to include. A match happens if the end of the filename matches a token in this list.",
        default=None,
    )
    exclude_files: Union[None, List[str]] = Field(
        description="List of file types (strings) to exclude. A match happens if the end of the filename matches a token in this list.",
        default=None,
    )
    exclude_dirs: Union[None, List[str]] = Field(
        description="Exclude directories that end with one of those tokens (list of strings).",
        default=None,
    )


class OnboardFileBundleModule(KiaraModule):
    """A generic module that imports a file from one of several possible sources."""

    _config_cls = OnboardFileBundleConfig

    @abc.abstractmethod
    def create_onboard_inputs_schema(self) -> Dict[str, Any]:
        pass

    def create_inputs_schema(
        self,
    ) -> ValueMapSchema:

        result = self.create_onboard_inputs_schema()

        for forbidden in [
            "bundle_name",
            "attach_metadata_to_bundle",
            "attach_metadata_to_files",
            "sub_path",
            "include_files",
            "exclude_files",
            "exclude_dirs",
        ]:
            if forbidden in result.keys():
                raise KiaraException(
                    msg=f"The '{forbidden}' input is not allowed in the onboard inputs schema for an implementing class of 'OnboardFileBundleModule'."
                )

        if self.get_config_value("result_bundle_name") is None:
            result["bundle_name"] = {
                "type": "string",
                "doc": "The bundle name use for the downloaded file, if not provided it will be autogenerated.",
                "optional": True,
            }

        if self.get_config_value("attach_metadata_to_bundle") is None:
            result["attach_metadata_to_bundle"] = {
                "type": "boolean",
                "doc": "Whether to attach the download metadata to the result file bundle instance.",
                "default": True,
            }
        if self.get_config_value("attach_metadata_to_files") is None:
            result["attach_metadata_to_bundle"] = {
                "type": "boolean",
                "doc": "Whether to attach the download metadata to each file in the resulting bundle.",
                "default": False,
            }

        if self.get_config_value("sub_path") is None:
            result["sub_path"] = {
                "type": "string",
                "doc": "The sub path to use. If not specified, the root of the source folder will be used.",
                "optional": True,
            }
        if self.get_config_value("include_files") is None:
            result["include_files"] = {
                "type": "list",
                "doc": "Include files that end with one of those tokens. If not specified, all file extensions are included.",
                "optional": True,
            }

        if self.get_config_value("exclude_files") is None:
            result["exclude_files"] = {
                "type": "list",
                "doc": "Exclude files that end with one of those tokens. If not specified, no file extensions are excluded.",
                "optional": True,
            }
        if self.get_config_value("exclude_dirs") is None:
            result["exclude_dirs"] = {
                "type": "list",
                "doc": "Exclude directories that end with one of those tokens. If not specified, no directories are excluded.",
                "optional": True,
            }

        return result

    def create_outputs_schema(
        self,
    ) -> ValueMapSchema:

        result = {
            "file_bundle": {
                "type": "file_bundle",
                "doc": "The file_bundle that was onboarded.",
            }
        }
        return result

    def process(self, inputs: ValueMap, outputs: ValueMap):

        from kiara.models.filesystem import FolderImportConfig, KiaraFileBundle

        bundle_name = self.get_config_value("result_bundle_name")
        if bundle_name is None:
            bundle_name = inputs.get_value_data("bundle_name")

        sub_path = self.get_config_value("sub_path")
        if sub_path is None:
            sub_path = inputs.get_value_data("sub_path")

        include = self.get_config_value("include_files")
        if include is None:
            _include = inputs.get_value_data("include_files")
            if _include:
                include = _include.list_data
        exclude = self.get_config_value("exclude_files")
        if exclude is None:
            _exclude = inputs.get_value_data("exclude_files")
            if _exclude:
                exclude = _exclude.list_data
        exclude_dirs = self.get_config_value("exclude_dirs")
        if exclude_dirs is None:
            _exclude_dirs = inputs.get_value_data("exclude_dirs")
            if _exclude_dirs:
                exclude_dirs = _exclude_dirs.list_data

        import_config_data = {
            "sub_path": sub_path,
        }
        if include:
            import_config_data["include_files"] = include
        if exclude:
            import_config_data["exclude_files"] = exclude
        if exclude_dirs:
            import_config_data["exclude_dirs"] = exclude_dirs

        import_config = FolderImportConfig(**import_config_data)

        attach_metadata_to_bundle = self.get_config_value("attach_metadata_to_bundle")
        if attach_metadata_to_bundle is None:
            attach_metadata_to_bundle = inputs.get_value_data(
                "attach_metadata_to_bundle"
            )

        attach_metadata_to_files = self.get_config_value("attach_metadata_to_files")
        if attach_metadata_to_files is None:
            attach_metadata_to_files = inputs.get_value_data("attach_metadata_to_files")

        archive = self.retrieve_archive(
            inputs=inputs,
            bundle_name=bundle_name,
            attach_metadata_to_bundle=attach_metadata_to_bundle,
            attach_metadata_to_files=attach_metadata_to_files,
            import_config=import_config,
        )
        if isinstance(archive, KiaraFileBundle):
            result = archive
        else:
            result = self.extract_archive(
                archive_file=archive,
                bundle_name=bundle_name,
                attach_metadata_to_bundle=attach_metadata_to_bundle,
                attach_metadata_to_files=attach_metadata_to_files,
                import_config=import_config,
            )

        outputs.set_value("file_bundle", result)

    @abc.abstractmethod
    def retrieve_archive(
        self,
        inputs: ValueMap,
        bundle_name: Union[str, None],
        attach_metadata_to_bundle: bool,
        attach_metadata_to_files: bool,
        import_config: "FolderImportConfig",
    ) -> Union["KiaraFile", "KiaraFileBundle"]:
        """Retrieve an archive file, or the actual result file bundle."""

    def extract_archive(
        self,
        archive_file: "KiaraFile",
        bundle_name: Union[str, None],
        attach_metadata_to_bundle: bool,
        attach_metadata_to_files: bool,
        import_config: "FolderImportConfig",
    ) -> "KiaraFileBundle":
        """Extract the archive file that was returned in 'retrieve_archive'."""

        from kiara.models.filesystem import KiaraFileBundle

        out_dir = tempfile.mkdtemp()

        def del_out_dir():
            shutil.rmtree(out_dir, ignore_errors=True)

        atexit.register(del_out_dir)

        error = None
        try:
            shutil.unpack_archive(archive_file.path, out_dir)
        except Exception:
            # try patool, maybe we're lucky
            try:
                import patoolib

                patoolib.extract_archive(archive_file.path, outdir=out_dir)
            except Exception as e:
                error = e

        if error is not None:
            raise KiaraProcessingException(f"Could not extract archive: {error}.")

        path = out_dir

        if not bundle_name:
            bundle_name = archive_file.file_name
            if import_config.sub_path:
                bundle_name = f"{bundle_name}#{import_config.sub_path}"

        bundle = KiaraFileBundle.import_folder(
            path, bundle_name=bundle_name, import_config=import_config
        )

        if attach_metadata_to_bundle:
            metadata = archive_file.metadata["download_info"]
            bundle.metadata["download_info"] = metadata

        if attach_metadata_to_files or True:
            metadata = archive_file.metadata["download_info"]
            for kf in bundle.included_files.values():
                kf.metadata["download_info"] = metadata

        return bundle



# -*- coding: utf-8 -*-
import atexit
import os
import tempfile
from datetime import datetime
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Mapping, Tuple, Type, Union

from pydantic import BaseModel, Field

from kiara.exceptions import KiaraException
from kiara.models.filesystem import FolderImportConfig, KiaraFile, KiaraFileBundle
from kiara.utils.dates import get_current_time_incl_timezone
from kiara.utils.files import unpack_archive
from kiara.utils.json import orjson_dumps
from kiara_plugin.onboarding.models import OnboardDataModel


class DownloadMetadata(BaseModel):
    url: str = Field(description="The url of the download request.")
    response_headers: List[Dict[str, str]] = Field(
        description="The response headers of the download request."
    )
    request_time: datetime = Field(description="The time the request was made.")
    download_time_in_seconds: float = Field(
        description="How long the download took in seconds."
    )


class DownloadBundleMetadata(DownloadMetadata):
    import_config: FolderImportConfig = Field(
        description="The import configuration that was used to import the files from the source bundle."
    )


@lru_cache()
def get_onboard_model_cls(
    onboard_type: Union[str, None]
) -> Union[None, Type[OnboardDataModel]]:

    if not onboard_type:
        return None

    from kiara.registries.models import ModelRegistry

    model_registry = ModelRegistry.instance()
    model_cls = model_registry.get_model_cls(onboard_type, OnboardDataModel)
    return model_cls  # type: ignore


def download_file(
    url: str,
    target: Union[str, None] = None,
    file_name: Union[str, None] = None,
    attach_metadata: bool = True,
    return_md5_hash: bool = False,
) -> Union[KiaraFile, Tuple[KiaraFile, str]]:

    import hashlib

    import httpx

    if not file_name:
        # TODO: make this smarter, using content-disposition headers if available
        file_name = url.split("/")[-1]

    if not target:
        tmp_file = tempfile.NamedTemporaryFile(delete=False, suffix=file_name)

        def rm_tmp_file():
            tmp_file.close()
            os.unlink(tmp_file.name)

        atexit.register(rm_tmp_file)

        _target = Path(tmp_file.name)
    else:
        _target = Path(target)
        _target.parent.mkdir(parents=True, exist_ok=True)

    if return_md5_hash:
        hash_md5 = hashlib.md5()  # noqa

    history = []

    request_time = get_current_time_incl_timezone()

    with open(_target, "wb") as f:
        with httpx.stream("GET", url, follow_redirects=True) as r:
            if r.status_code < 200 or r.status_code >= 399:
                raise KiaraException(
                    f"Could not download file from {url}: status code {r.status_code}."
                )
            history.append(dict(r.headers))
            for h in r.history:
                history.append(dict(h.headers))
            for data in r.iter_bytes():
                if return_md5_hash:
                    hash_md5.update(data)
                f.write(data)

    result_file = KiaraFile.load_file(_target.as_posix(), file_name)
    now_time = get_current_time_incl_timezone()
    delta = (now_time - request_time).total_seconds()
    if attach_metadata:
        metadata = {
            "url": url,
            "response_headers": history,
            "request_time": request_time,
            "download_time_in_seconds": delta,
        }
        _metadata: DownloadMetadata = DownloadMetadata(**metadata)
        result_file.metadata["download_info"] = _metadata.model_dump()
        result_file.metadata_schemas["download_info"] = orjson_dumps(
            DownloadMetadata.model_json_schema()
        )

    if return_md5_hash:
        return result_file, hash_md5.hexdigest()
    else:
        return result_file


def download_file_bundle(
    url: str,
    attach_metadata: bool = True,
    import_config: Union[FolderImportConfig, None] = None,
) -> KiaraFileBundle:

    import shutil
    from datetime import datetime
    from urllib.parse import urlparse

    import httpx
    import pytz

    suffix = None
    try:
        parsed_url = urlparse(url)
        _, suffix = os.path.splitext(parsed_url.path)
    except Exception:
        pass
    if not suffix:
        suffix = ""

    tmp_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)

    def rm_tmp_file():
        tmp_file.close()
        os.unlink(tmp_file.name)

    atexit.register(rm_tmp_file)

    history = []
    # datetime.utcnow().replace(tzinfo=pytz.utc)
    with open(tmp_file.name, "wb") as f:
        with httpx.stream("GET", url, follow_redirects=True) as r:
            history.append(dict(r.headers))
            for h in r.history:
                history.append(dict(h.headers))
            for data in r.iter_bytes():
                f.write(data)

    out_dir = tempfile.mkdtemp()

    def del_out_dir():
        shutil.rmtree(out_dir, ignore_errors=True)

    atexit.register(del_out_dir)

    # error = None
    # try:
    #     shutil.unpack_archive(tmp_file.name, out_dir)
    # except Exception:
    #     # try patool, maybe we're lucky
    #     try:
    #         import patoolib
    #
    #         patoolib.extract_archive(tmp_file.name, outdir=out_dir)
    #     except Exception as e:
    #         error = e
    #
    # if error is not None:
    #     raise KiaraException(msg=f"Could not extract archive: {error}.")

    unpack_archive(tmp_file.name, out_dir)
    bundle = KiaraFileBundle.import_folder(out_dir, import_config=import_config)

    if import_config is None:
        ic_dict = {}
    elif isinstance(import_config, FolderImportConfig):
        ic_dict = import_config.dict()
    else:
        ic_dict = import_config
    if attach_metadata:
        metadata = {
            "url": url,
            "response_headers": history,
            "request_time": datetime.utcnow().replace(tzinfo=pytz.utc).isoformat(),
            "import_config": ic_dict,
        }
        _metadata = DownloadBundleMetadata(**metadata)
        bundle.metadata["download_info"] = _metadata.model_dump()
        bundle.metadata_schemas["download_info"] = DownloadMetadata.schema_json()

    return bundle


def find_matching_onboard_models(
    uri: str, for_bundle: bool = False
) -> Mapping[Type[OnboardDataModel], Tuple[bool, str]]:

    from kiara.registries.models import ModelRegistry

    model_registry = ModelRegistry.instance()
    onboard_models = model_registry.get_models_of_type(
        OnboardDataModel
    ).item_infos.values()

    result = {}
    onboard_model: Type[OnboardDataModel]
    for onboard_model in onboard_models:  # type: ignore

        python_cls: Type[OnboardDataModel] = onboard_model.python_class.get_class()  # type: ignore
        if for_bundle:
            result[python_cls] = python_cls.accepts_bundle_uri(uri)
        else:
            result[python_cls] = python_cls.accepts_uri(uri)

    return result


def onboard_file(
    source: str,
    file_name: Union[str, None] = None,
    onboard_type: Union[str, None] = None,
    attach_metadata: bool = True,
) -> KiaraFile:

    if not onboard_type:

        model_clsses = find_matching_onboard_models(source)
        matches = [k for k, v in model_clsses.items() if v[0]]
        if not matches:
            raise KiaraException(
                msg=f"Can't onboard file from '{source}': no onboard models found that accept this source type."
            )
        elif len(matches) > 1:
            msg = "Valid onboarding types for this uri:\n\n"
            for k, v in model_clsses.items():
                if not v[0]:
                    continue
                msg += f"  - {k._kiara_model_id}: {v[1]}\n"
            raise KiaraException(
                msg=f"Can't onboard file from '{source}': multiple onboard models found that accept this source type.\n\n{msg}"
            )

        model_cls: Type[OnboardDataModel] = matches[0]

    else:

        model_cls = get_onboard_model_cls(onboard_type=onboard_type)  # type: ignore
        if not model_cls:
            raise KiaraException(msg=f"Can't onboard file from '{source}' using onboard type '{onboard_type}': no onboard model found with this name.")  # type: ignore

        valid, msg = model_cls.accepts_uri(source)
        if not valid:
            raise KiaraException(msg=f"Can't onboard file from '{source}' using onboard type '{model_cls._kiara_model_id}': {msg}")  # type: ignore

    if not model_cls.get_config_fields():
        model = model_cls()
    else:
        raise NotImplementedError()

    result = model.retrieve(
        uri=source, file_name=file_name, attach_metadata=attach_metadata
    )
    if not result:
        raise KiaraException(msg=f"Can't onboard file from '{source}' using onboard type '{model_cls._kiara_model_id}': no result data retrieved. This is most likely a bug.")  # type: ignore

    if isinstance(result, str):
        data = KiaraFile.load_file(result, file_name=file_name)
    elif not isinstance(result, KiaraFile):
        raise KiaraException(
            "Can't onboard file: onboard model returned data that is not a file. This is most likely a bug."
        )
    else:
        data = result

    return data


def onboard_file_bundle(
    source: str,
    import_config: Union[FolderImportConfig, None],
    onboard_type: Union[str, None] = None,
    attach_metadata: bool = True,
) -> KiaraFileBundle:

    if not onboard_type:

        model_clsses = find_matching_onboard_models(uri=source, for_bundle=True)
        matches = [k for k, v in model_clsses.items() if v[0]]
        if not matches:
            raise KiaraException(
                msg=f"Can't onboard file from '{source}': no onboard models found that accept this source type."
            )
        elif len(matches) > 1:
            msg = "Valid onboarding types for this uri:\n\n"
            for k, v in model_clsses.items():
                if not v[0]:
                    continue
                msg += f"  - {k._kiara_model_id}: {v[1]}\n"
            raise KiaraException(
                msg=f"Can't onboard file from '{source}': multiple onboard models found that accept this source type.\n\n{msg}"
            )

        model_cls: Type[OnboardDataModel] = matches[0]

    else:
        model_cls = get_onboard_model_cls(onboard_type=onboard_type)  # type: ignore
        if not model_cls:
            raise KiaraException(msg=f"Can't onboard file from '{source}' using onboard type '{onboard_type}': no onboard model found with this name.")  # type: ignore
        valid, msg = model_cls.accepts_bundle_uri(source)
        if not valid:
            raise KiaraException(msg=f"Can't onboard file from '{source}' using onboard type '{model_cls._kiara_model_id}': {msg}")  # type: ignore

    if not model_cls.get_config_fields():
        model = model_cls()
    else:
        raise NotImplementedError()

    if not import_config:
        import_config = FolderImportConfig()

    try:
        result: Union[None, KiaraFileBundle] = model.retrieve_bundle(
            uri=source, import_config=import_config, attach_metadata=attach_metadata
        )

        if not result:
            raise KiaraException(msg=f"Can't onboard file bundle from '{source}' using onboard type '{model_cls._kiara_model_id}': no result data retrieved. This is most likely a bug.")  # type: ignore

        if isinstance(result, str):
            result = KiaraFileBundle.import_folder(source=result)

    except NotImplementedError:
        result = None

    if not result:
        result_file = model.retrieve(
            uri=source, file_name=None, attach_metadata=attach_metadata
        )
        if not result_file:
            raise KiaraException(msg=f"Can't onboard file bundle from '{source}' using onboard type '{model_cls._kiara_model_id}': no result data retrieved. This is most likely a bug.")  # type: ignore

        if isinstance(result, str):
            imported_bundle_file = KiaraFile.load_file(result_file)  # type: ignore
        elif not isinstance(result_file, KiaraFile):
            raise KiaraException(
                "Can't onboard file: onboard model returned data that is not a file. This is most likely a bug."
            )
        else:
            imported_bundle_file = result_file

        imported_bundle = KiaraFileBundle.from_archive_file(
            imported_bundle_file, import_config=import_config
        )
    else:
        imported_bundle = result

    return imported_bundle


def download_zenodo_file_bundle(
    doi: str,
    version: Union[None, str],
    attach_metadata_to_bundle: bool,
    attach_metadata_to_files: bool,
    bundle_name: Union[str, None] = None,
    import_config: Union[None, Mapping[str, Any], FolderImportConfig] = None,
) -> KiaraFileBundle:

    import pyzenodo3

    from kiara.models.filesystem import KiaraFile, KiaraFileBundle

    if "/zenodo." not in doi:
        doi = f"10.5281/zenodo.{doi}"

    zen = pyzenodo3.Zenodo()

    if version:
        raise NotImplementedError("Downloading versioned records is not yet supported.")

    record = zen.find_record_by_doi(doi)

    base_path = KiaraFileBundle.create_tmp_dir()

    for _available_file in record.data["files"]:
        match = _available_file

        url = match["links"]["self"]
        checksum = match["checksum"][4:]

        file_path = _available_file["key"]
        full_path = base_path / file_path

        file_name = file_path.split("/")[-1]

        # TODO: filter here already, so we don't need to download files we don't want

        result_file: KiaraFile
        result_file, result_checksum = download_file(  # type: ignore
            url=url,
            target=full_path.as_posix(),
            file_name=file_name,
            attach_metadata=True,
            return_md5_hash=True,
        )

        if checksum != result_checksum:
            raise KiaraException(
                msg=f"Can't download file '{file_name}' from zenodo, invalid checksum: {checksum} != {result_checksum}"
            )

    if not bundle_name:
        bundle_name = doi
    result = KiaraFileBundle.import_folder(
        source=base_path.as_posix(),
        bundle_name=bundle_name,
        import_config=import_config,
    )
    if attach_metadata_to_bundle:
        result.metadata["zenodo_record_data"] = record.data

    if attach_metadata_to_files:
        for file in result.included_files.values():
            file.metadata["zenodo_record_data"] = record.data

    return result



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, Any, Dict, Union

from kiara.api import ValueMap
from kiara.exceptions import KiaraException
from kiara_plugin.onboarding.modules import OnboardFileBundleModule, OnboardFileModule

if TYPE_CHECKING:
    from kiara.models.filesystem import FolderImportConfig, KiaraFile, KiaraFileBundle


class DownloadZenodoFileModule(OnboardFileModule):
    """Download a single file from a Zenodo record."""

    _module_type_name = "download.file.from.zenodo"

    def create_onboard_inputs_schema(self) -> Dict[str, Any]:

        result: Dict[str, Dict[str, Any]] = {
            "doi": {"type": "string", "doc": "The DOI."},
            "version": {
                "type": "string",
                "doc": "The version of the record to download.",
                "optional": True,
            },
            "path": {
                "type": "string",
                "doc": "The path to the file/file name within the dataset.",
                "optional": True,
            },
        }
        return result

    def retrieve_file(
        self, inputs: ValueMap, file_name: Union[str, None], attach_metadata: bool
    ) -> Any:

        import pyzenodo3

        from kiara_plugin.onboarding.utils.download import download_file

        doi = inputs.get_value_data("doi")

        version = inputs.get_value_data("version")
        if version:
            raise NotImplementedError(
                "Downloading versioned records is not yet supported."
            )

        file_path = inputs.get_value_data("path")

        if "/zenodo." not in doi:
            doi = f"10.5281/zenodo.{doi}"

        zen = pyzenodo3.Zenodo()
        record = zen.find_record_by_doi(doi)

        if not file_path:
            if len(record.data["files"]) == 1:
                file_path = record.data["files"][0]["key"]
            else:
                msg = "Available files:\n"
                for key in record.data["files"]:
                    msg += f"  - {key['key']}\n"

                raise KiaraException(
                    msg=f"Multiple files available in Zenodo record, please specify 'path' input.\n\n{msg}"
                )

        match = None
        for _available_file in record.data["files"]:
            if file_path == _available_file["key"]:
                match = _available_file
                break

        if not match:
            msg = "Available files:\n"
            for key in record.data["files"]:
                msg += f"  - {key['key']}\n"
            raise KiaraException(
                msg=f"Can't find file '{file_path}' in Zenodo record. {msg}"
            )

        url = match["links"]["self"]
        checksum = match["checksum"][4:]

        file_name = file_path.split("/")[-1]

        result_file: KiaraFile
        result_file, result_checksum = download_file(  # type: ignore
            url=url,
            file_name=file_name,
            attach_metadata=attach_metadata,
            return_md5_hash=True,
        )

        if checksum != result_checksum:
            raise KiaraException(
                msg=f"Can't download file '{file_name}' from zenodo, invalid checksum: {checksum} != {checksum}"
            )

        if attach_metadata:
            result_file.metadata["zenodo_record_data"] = record.data

        return result_file


class DownloadZenodoFileBundleModule(OnboardFileBundleModule):
    """Download a file bundle from a remote github repository.

    If 'sub_path' is not specified, the whole repo will be used.

    """

    _module_type_name = "download.file_bundle.from.zenodo"

    def create_onboard_inputs_schema(self) -> Dict[str, Any]:
        result: Dict[str, Dict[str, Any]] = {
            "doi": {"type": "string", "doc": "The DOI."},
            "version": {
                "type": "string",
                "doc": "The version of the record to download. By default, the latest version will be used.",
                "optional": True,
            },
        }
        return result

    def retrieve_archive(
        self,
        inputs: ValueMap,
        bundle_name: Union[str, None],
        attach_metadata_to_bundle: bool,
        attach_metadata_to_files: bool,
        import_config: "FolderImportConfig",
    ) -> Union["KiaraFile", "KiaraFileBundle"]:

        from kiara_plugin.onboarding.utils.download import download_zenodo_file_bundle

        doi = inputs.get_value_data("doi")
        version = inputs.get_value_data("version")

        result = download_zenodo_file_bundle(
            doi=doi,
            version=version,
            attach_metadata_to_bundle=attach_metadata_to_bundle,
            attach_metadata_to_files=attach_metadata_to_files,
            bundle_name=bundle_name,
            import_config=import_config,
        )
        return result



# -*- coding: utf-8 -*-
import os

from kiara.models.filesystem import KiaraFileBundle
from kiara.models.values.value import Value


def check_downloaded_file(file_bundle: Value):

    assert file_bundle.data_type_name == "file_bundle"
    assert file_bundle.value_size > 400000

    kiara_file_bundle: KiaraFileBundle = file_bundle.data
    assert kiara_file_bundle.__class__ == KiaraFileBundle

    assert (
        f"kiara_plugin.core_types-develop{os.path.sep}.gitignore"
        in kiara_file_bundle.included_files.keys()
    )



# -*- coding: utf-8 -*-
import atexit
import os
import tempfile
from typing import TYPE_CHECKING, Any, Dict, Union

from kiara.models.values.value import ValueMap
from kiara_plugin.onboarding.modules import OnboardFileBundleModule, OnboardFileModule

if TYPE_CHECKING:
    from kiara.models.filesystem import FolderImportConfig, KiaraFile, KiaraFileBundle


class DownloadFileModule(OnboardFileModule):
    """Download a single file from a remote location.

    The result of this operation is a single value of type 'file' (basically an array of raw bytes + some light metadata), which can then be used in other modules to create more meaningful data structures.
    """

    _module_type_name = "download.file"

    def create_onboard_inputs_schema(self) -> Dict[str, Any]:

        result: Dict[str, Dict[str, Any]] = {
            "url": {"type": "string", "doc": "The url of the file to download."},
        }
        return result

    def retrieve_file(
        self, inputs: ValueMap, file_name: Union[str, None], attach_metadata: bool
    ) -> Any:

        from kiara_plugin.onboarding.utils.download import download_file

        url = inputs.get_value_data("url")

        result_file = download_file(
            url=url,
            file_name=file_name,
            attach_metadata=attach_metadata,
        )
        return result_file


class DownloadFileBundleModule(OnboardFileBundleModule):
    """Download a file bundle from a remote location.

    This is basically just a convenience module that incorporates unpacking of the downloaded file into a folder structure, and then wrapping it into a *kiara* `file_bundle` data type.

    If the `sub_path` input is set, the whole data is downloaded anyway, but before wrapping into a `file_bundle` value, the files not in the sub-path are ignored (and thus not available later on). Make sure you
    decided whether this is ok for your use-case, if not, rather filter the `file_bundle` later in an
    extra step (for example using the `file_bundle.pick.sub_folder` operation).
    """

    _module_type_name = "download.file_bundle"

    def create_onboard_inputs_schema(self) -> Dict[str, Any]:

        result: Dict[str, Dict[str, Any]] = {
            "url": {
                "type": "string",
                "doc": "The url of an archive/zip file to download.",
            }
        }

        return result

    def retrieve_archive(
        self,
        inputs: ValueMap,
        bundle_name: Union[str, None],
        attach_metadata_to_bundle: bool,
        attach_metadata_to_files: bool,
        import_config: "FolderImportConfig",
    ) -> Union["KiaraFile", "KiaraFileBundle"]:

        from urllib.parse import urlparse

        from kiara.models.filesystem import KiaraFile
        from kiara_plugin.onboarding.utils.download import download_file

        url = inputs.get_value_data("url")
        suffix = None
        try:
            parsed_url = urlparse(url)
            _, suffix = os.path.splitext(parsed_url.path)
        except Exception:
            pass
        if not suffix:
            suffix = ""

        tmp_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)

        def rm_tmp_file():
            tmp_file.close()
            os.unlink(tmp_file.name)

        atexit.register(rm_tmp_file)
        kiara_file: KiaraFile

        kiara_file = download_file(  # type: ignore
            url, target=tmp_file.name, attach_metadata=True, return_md5_hash=False
        )

        assert kiara_file.path == tmp_file.name

        return kiara_file



# -*- coding: utf-8 -*-


# class ImportFileBundleConfig(KiaraModuleConfig):
#
#     import_metadata: bool = Field(
#         description="Whether to return the import metadata as well.",
#         default=True,
#     )
#
#
# class ImportFileBundleModule(KiaraModule):
#     """A generic module to import a file bundle from any local or remote location."""
#
#     _module_type_name = "import.file_bundle"
#     _config_cls = ImportFileBundleConfig
#
#     def create_inputs_schema(
#         self,
#     ) -> ValueMapSchema:
#
#         result: Dict[str, Dict[str, Any]] = {
#             "uri": {
#                 "type": "string",
#                 "doc": "The uri (url/path/...) of the file to import.",
#             }
#         }
#         if self.get_config_value("import_metadata"):
#             result["import_metadata"] = {
#                 "type": "dict",
#                 "doc": "Metadata you want to attach to the file bundle.",
#                 "optional": True,
#             }
#
#         return result
#
#     def create_outputs_schema(
#         self,
#     ) -> ValueMapSchema:
#
#         result = {
#             "file_bundle": {
#                 "type": "file_bundle",
#                 "doc": "The imported file bundle.",
#             }
#         }
#         if self.get_config_value("import_metadata"):
#             result["import_metadata"] = {
#                 "type": "dict",
#                 "doc": "Metadata about the import and file bundle.",
#             }
#         return result
#
#     def process(self, inputs: ValueMap, outputs: ValueMap) -> None:
#         pass



# -*- coding: utf-8 -*-
# import atexit
# import os
# import shutil
# import tempfile
# from typing import Any, Dict, Union
#
# from pydantic import Field
#
# from kiara.api import KiaraModule, KiaraModuleConfig, ValueMap, ValueMapSchema
# from kiara.exceptions import KiaraProcessingException
# from kiara.models.filesystem import FileBundle
# from kiara_plugin.onboarding.utils.download import download_file
#
#
# class ImportFileConfig(KiaraModuleConfig):
#
#     import_metadata: bool = Field(
#         description="Whether to return the import metadata as well.",
#         default=True,
#     )
#     source_location: Union[str, None] = Field(
#         description="The location of the source file.", default=None
#     )
#
#
# class DownloadFileConfig(KiaraModuleConfig):
#     attach_metadata: bool = Field(
#         description="Whether to attach the download metadata to the result file.",
#         default=True,
#     )
#
#
# class DownloadFileModule(KiaraModule):
#     """Download a single file from a remote location.
#
#     The result of this operation is a single value of type 'file' (basically an array of raw bytes), which can then be used in other modules to
#     create more meaningful data structures.
#     """
#
#     _module_type_name = "download.file"
#     _config_cls = DownloadFileConfig
#
#     def create_inputs_schema(self) -> ValueMapSchema:
#
#         result: Dict[str, Dict[str, Any]] = {
#             "url": {"type": "string", "doc": "The url of the file to download."},
#             "file_name": {
#                 "type": "string",
#                 "doc": "The file name to use for the downloaded file.",
#                 "optional": True,
#             },
#         }
#         return result
#
#     def create_outputs_schema(
#         self,
#     ) -> ValueMapSchema:
#
#         result: Dict[str, Dict[str, Any]] = {
#             "file": {
#                 "type": "file",
#                 "doc": "The downloaded file.",
#             }
#         }
#
#         return result
#
#     def process(self, inputs: ValueMap, outputs: ValueMap):
#
#         url = inputs.get_value_data("url")
#         file_name = inputs.get_value_data("file_name")
#
#         result_file = download_file(
#             url=url,
#             file_name=file_name,
#             attach_metadata=self.get_config_value("attach_metadata"),
#         )
#
#         outputs.set_value("file", result_file)
#
#
# class DownloadFileBundleModule(KiaraModule):
#     _module_type_name = "download.file_bundle"
#     _config_cls = DownloadFileConfig
#
#     def create_inputs_schema(self) -> ValueMapSchema:
#
#         result: Dict[str, Dict[str, Any]] = {
#             "url": {
#                 "type": "string",
#                 "doc": "The url of an archive/zip file to download.",
#             },
#             "sub_path": {
#                 "type": "string",
#                 "doc": "A relative path to select only a sub-folder from the archive.",
#                 "optional": True,
#             },
#         }
#
#         return result
#
#     def create_outputs_schema(
#         self,
#     ) -> ValueMapSchema:
#
#         result: Dict[str, Dict[str, Any]] = {
#             "file_bundle": {
#                 "type": "file_bundle",
#                 "doc": "The downloaded file bundle.",
#             }
#         }
#
#         return result
#
#     def process(self, inputs: ValueMap, outputs: ValueMap):
#
#         from datetime import datetime
#         from urllib.parse import urlparse
#
#         import httpx
#         import pytz
#
#         url = inputs.get_value_data("url")
#         suffix = None
#         try:
#             parsed_url = urlparse(url)
#             _, suffix = os.path.splitext(parsed_url.path)
#         except Exception:
#             pass
#         if not suffix:
#             suffix = ""
#
#         sub_path: Union[None, str] = inputs.get_value_data("sub_path")
#         tmp_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)
#         atexit.register(tmp_file.close)
#
#         history = []
#         datetime.utcnow().replace(tzinfo=pytz.utc)
#         with open(tmp_file.name, "wb") as f:
#             with httpx.stream("GET", url, follow_redirects=True) as r:
#                 history.append(dict(r.headers))
#                 for h in r.history:
#                     history.append(dict(h.headers))
#                 for data in r.iter_bytes():
#                     f.write(data)
#
#         out_dir = tempfile.mkdtemp()
#
#         def del_out_dir():
#             shutil.rmtree(out_dir, ignore_errors=True)
#
#         atexit.register(del_out_dir)
#
#         error = None
#         try:
#             shutil.unpack_archive(tmp_file.name, out_dir)
#         except Exception:
#             # try patool, maybe we're lucky
#             try:
#                 import patoolib
#
#                 patoolib.extract_archive(tmp_file.name, outdir=out_dir)
#             except Exception as e:
#                 error = e
#
#         if error is not None:
#             raise KiaraProcessingException(f"Could not extract archive: {error}.")
#
#         path = out_dir
#         if sub_path:
#             path = os.path.join(out_dir, sub_path)
#         bundle = FileBundle.import_folder(path)
#
#         metadata = {
#             "response_headers": history,
#             "request_time": datetime.utcnow().replace(tzinfo=pytz.utc).isoformat(),
#         }
#         outputs.set_value("download_metadata", metadata)
#         outputs.set_value("file_bundle", bundle)



# -*- coding: utf-8 -*-

"""This module contains the metadata (and other) models that are used in the ``kiara_plugin.onboarding`` package.

Those models are convenience wrappers that make it easier for *kiara* to find, create, manage and version metadata -- but also
other type of models -- that is attached to data, as well as *kiara* modules.

Metadata models must be a sub-class of [kiara.metadata.MetadataModel][kiara.metadata.MetadataModel]. Other models usually
sub-class a pydantic BaseModel or implement custom base classes.
"""
import os.path
from abc import abstractmethod
from typing import ClassVar, List, Tuple, Union

from kiara.exceptions import KiaraException
from kiara.models import KiaraModel
from kiara.models.filesystem import FolderImportConfig, KiaraFile, KiaraFileBundle


class OnboardDataModel(KiaraModel):

    _kiara_model_id: ClassVar[str] = None  # type: ignore

    @classmethod
    def get_config_fields(cls) -> List[str]:
        return sorted(cls.model_fields.keys())

    @classmethod
    @abstractmethod
    def accepts_uri(cls, uri: str) -> Tuple[bool, str]:
        pass

    @classmethod
    def accepts_bundle_uri(cls, uri: str) -> Tuple[bool, str]:
        return cls.accepts_uri(uri)

    @abstractmethod
    def retrieve(
        self, uri: str, file_name: Union[None, str], attach_metadata: bool
    ) -> KiaraFile:
        pass

    def retrieve_bundle(
        self, uri: str, import_config: FolderImportConfig, attach_metadata: bool
    ) -> KiaraFileBundle:
        raise NotImplementedError()


class FileFromLocalModel(OnboardDataModel):

    _kiara_model_id: ClassVar[str] = "onboarding.file.from.local_file"

    @classmethod
    def accepts_uri(cls, uri: str) -> Tuple[bool, str]:

        if os.path.isfile(os.path.abspath(uri)):
            return True, "local file exists and is file"
        else:
            return False, "local file does not exist or is not a file"

    @classmethod
    def accepts_bundle_uri(cls, uri: str) -> Tuple[bool, str]:

        if os.path.isdir(os.path.abspath(uri)):
            return True, "local folder exists and is folder"
        else:
            return False, "local folder does not exist or is not a folder"

    def retrieve(
        self, uri: str, file_name: Union[None, str], attach_metadata: bool
    ) -> KiaraFile:

        if not os.path.exists(os.path.abspath(uri)):
            raise KiaraException(
                f"Can't create file from path '{uri}': path does not exist."
            )
        if not os.path.isfile(os.path.abspath(uri)):
            raise KiaraException(
                f"Can't create file from path '{uri}': path is not a file."
            )

        return KiaraFile.load_file(uri)

    def retrieve_bundle(
        self, uri: str, import_config: FolderImportConfig, attach_metadata: bool
    ) -> KiaraFileBundle:

        if not os.path.exists(os.path.abspath(uri)):
            raise KiaraException(
                f"Can't create file from path '{uri}': path does not exist."
            )
        if not os.path.isdir(os.path.abspath(uri)):
            raise KiaraException(
                f"Can't create file from path '{uri}': path is not a directory."
            )

        return KiaraFileBundle.import_folder(source=uri, import_config=import_config)


class FileFromRemoteModel(OnboardDataModel):

    _kiara_model_id: ClassVar[str] = "onboarding.file.from.url"

    @classmethod
    def accepts_uri(cls, uri: str) -> Tuple[bool, str]:

        accepted_protocols = ["http", "https"]
        for protocol in accepted_protocols:
            if uri.startswith(f"{protocol}://"):
                return True, "url is valid (starts with http or https)"

        return False, "url is not valid (does not start with http or https)"

    def retrieve(
        self, uri: str, file_name: Union[None, str], attach_metadata: bool
    ) -> KiaraFile:
        from kiara_plugin.onboarding.utils.download import download_file

        result_file: KiaraFile = download_file(  # type: ignore
            url=uri, file_name=file_name, attach_metadata=attach_metadata
        )
        return result_file

    def retrieve_bundle(
        self, uri: str, import_config: FolderImportConfig, attach_metadata: bool
    ) -> KiaraFileBundle:
        from kiara_plugin.onboarding.utils.download import download_file_bundle

        result_bundle = download_file_bundle(
            url=uri, import_config=import_config, attach_metadata=attach_metadata
        )
        return result_bundle


class FileFromZenodoModel(OnboardDataModel):

    _kiara_model_id: ClassVar[str] = "onboarding.file.from.zenodo"

    @classmethod
    def accepts_uri(cls, uri: str) -> Tuple[bool, str]:

        if uri.startswith("zenodo:"):
            return True, "url is valid (follows format 'zenodo:<doi>')"

        elif "/zenodo." in uri:
            return True, "url is valid (contains '/zenodo.')"

        return False, "url is not valid (does not follow format 'zenodo:<doi>')"

    def retrieve(
        self, uri: str, file_name: Union[None, str], attach_metadata: bool
    ) -> KiaraFile:

        import pyzenodo3

        from kiara_plugin.onboarding.utils.download import download_file

        if uri.startswith("zenodo:"):
            doi = uri[len("zenodo:") :]
        elif "/zenodo." in uri:
            doi = uri

        tokens = doi.split("/zenodo.")
        if len(tokens) != 2:
            raise KiaraException(
                msg=f"Can't parse Zenodo DOI from URI for single file download: {doi}"
            )

        path_components = tokens[1].split("/", maxsplit=1)
        if len(path_components) != 2:
            raise KiaraException(
                msg=f"Can't parse Zenodo DOI from URI for single file download: {doi}"
            )

        file_path = path_components[1]
        _doi = f"{tokens[0]}/zenodo.{path_components[0]}"

        zen = pyzenodo3.Zenodo()
        record = zen.find_record_by_doi(_doi)

        match = None
        for _available_file in record.data["files"]:
            if file_path == _available_file["key"]:
                match = _available_file
                break

        if not match:
            msg = "Available files:\n"
            for key in record.data["files"]:
                msg += f"  - {key['key']}\n"
            raise KiaraException(
                msg=f"Can't find file '{file_path}' in Zenodo record. {msg}"
            )

        url = match["links"]["self"]
        checksum = match["checksum"][4:]

        file_name = file_path.split("/")[-1]

        file_model: KiaraFile
        file_model, md5_digest = download_file(  # type: ignore
            url=url,
            target=None,
            file_name=file_name,
            attach_metadata=attach_metadata,
            return_md5_hash=True,
        )

        if checksum != md5_digest:
            raise KiaraException(
                msg=f"Can't download file '{file_name}', invalid checksum: {checksum} != {md5_digest}"
            )

        if attach_metadata:
            file_model.metadata["zenodo_record_data"] = record.data

        return file_model

    def retrieve_bundle(
        self, uri: str, import_config: FolderImportConfig, attach_metadata: bool
    ) -> KiaraFileBundle:

        import shutil

        import pyzenodo3

        from kiara_plugin.onboarding.utils.download import download_file

        if uri.startswith("zenodo:"):
            doi = uri[len("zenodo:") :]
        elif "/zenodo." in uri:
            doi = uri

        tokens = doi.split("/zenodo.")
        if len(tokens) != 2:
            raise KiaraException(
                msg=f"Can't parse Zenodo DOI from URI for single file download: {doi}"
            )

        path_components = tokens[1].split("/", maxsplit=1)

        if len(path_components) == 2:
            zid = path_components[0]
            file_path = path_components[1]
        else:
            zid = path_components[0]
            file_path = None

        _doi = f"{tokens[0]}/zenodo.{zid}"

        if not file_path:

            zen = pyzenodo3.Zenodo()

            record = zen.find_record_by_doi(_doi)

            path = KiaraFileBundle.create_tmp_dir()
            shutil.rmtree(path, ignore_errors=True)
            path.mkdir()

            for file_data in record.data["files"]:
                url = file_data["links"]["self"]
                file_name = file_data["key"]
                checksum = file_data["checksum"][4:]

                target = os.path.join(path, file_name)
                file_model: KiaraFile
                file_model, md5_digest = download_file(  # type: ignore
                    url=url,
                    target=target,
                    file_name=file_name,
                    attach_metadata=attach_metadata,
                    return_md5_hash=True,
                )

                if checksum != md5_digest:
                    raise KiaraException(
                        msg=f"Can't download file '{file_name}', invalid checksum: {checksum} != {md5_digest}"
                    )

            bundle = KiaraFileBundle.import_folder(path.as_posix())
            if attach_metadata:
                bundle.metadata["zenodo_record_data"] = record.data

        else:

            zen = pyzenodo3.Zenodo()
            record = zen.find_record_by_doi(_doi)

            match = None
            for _available_file in record.data["files"]:
                if file_path == _available_file["key"]:
                    match = _available_file
                    break

            if not match:
                msg = "Available files:\n"
                for key in record.data["files"]:
                    msg += f"  - {key['key']}\n"
                raise KiaraException(
                    msg=f"Can't find file '{file_path}' in Zenodo record. {msg}"
                )

            url = match["links"]["self"]
            checksum = match["checksum"][4:]

            file_name = file_path.split("/")[-1]

            file_model, md5_digest = download_file(  # type: ignore
                url=url,
                target=None,
                file_name=file_name,
                attach_metadata=attach_metadata,
                return_md5_hash=True,
            )

            if checksum != md5_digest:
                raise KiaraException(
                    msg=f"Can't download file '{file_name}', invalid checksum: {checksum} != {md5_digest}"
                )

            bundle = KiaraFileBundle.from_archive_file(
                archive_file=file_model, import_config=import_config
            )
            if attach_metadata:
                bundle.metadata["zenodo_record_data"] = record.data

        return bundle


class FileFromZoteroModel(OnboardDataModel):

    _kiara_model_id: ClassVar[str] = "onboarding.file.from.zotero"

    @classmethod
    def accepts_uri(cls, uri: str) -> Tuple[bool, str]:
        if uri.startswith("zotero:"):
            return True, "uri is a zotero uri"
        else:
            return False, "uri is not a zotero uri, must start with 'zotero:'"


class FileFromGithubModel(OnboardDataModel):

    _kiara_model_id: ClassVar[str] = "onboarding.file.from.github"

    @classmethod
    def accepts_uri(cls, uri: str) -> Tuple[bool, str]:

        if uri.startswith("gh:") or uri.startswith("github:"):
            return True, "uri is a github uri"

        return False, "uri is not a github uri, must start with 'gh:' or 'github:'"

    def retrieve(
        self, uri: str, file_name: Union[None, str], attach_metadata: bool
    ) -> KiaraFile:
        from kiara_plugin.onboarding.utils.download import download_file

        tokens = uri.split(":")[1].split("/", maxsplit=3)
        if len(tokens) != 4:
            raise KiaraException(
                msg=f"Can't parse github uri '{uri}' for single file download. Required format: 'gh:<user>/<repo>/<branch_or_tag>/<path>'"
            )

        url = f"https://raw.githubusercontent.com/{tokens[0]}/{tokens[1]}/{tokens[2]}/{tokens[3]}"

        result_file: KiaraFile = download_file(  # type: ignore
            url=url, attach_metadata=attach_metadata
        )
        return result_file

    def retrieve_bundle(
        self, uri: str, import_config: FolderImportConfig, attach_metadata: bool
    ) -> KiaraFileBundle:

        from kiara_plugin.onboarding.utils.download import download_file

        tokens = uri.split(":")[1].split("/", maxsplit=3)
        if len(tokens) == 3:
            sub_path = None
        elif len(tokens) != 4:
            raise KiaraException(
                msg=f"Can't parse github uri '{uri}' for single file download. Required format: 'gh:<user>/<repo>/<branch_or_tag>/<path>'"
            )
        else:
            sub_path = tokens[3]

        url = f"https://github.com/{tokens[0]}/{tokens[1]}/archive/refs/heads/{tokens[2]}.zip"
        file_name = f"{tokens[1]}-{tokens[2]}.zip"

        archive_zip: KiaraFile
        archive_zip = download_file(  # type: ignore
            url=url,
            attach_metadata=attach_metadata,
            file_name=file_name,
            return_md5_hash=False,
        )

        base_sub_path = f"{tokens[1]}-{tokens[2]}"

        if sub_path:
            if import_config.sub_path:
                new_sub_path = "/".join([base_sub_path, sub_path, import_config.sub_path])  # type: ignore
            else:
                new_sub_path = "/".join([base_sub_path, sub_path])
        elif import_config.sub_path:
            new_sub_path = "/".join([base_sub_path, import_config.sub_path])
        else:
            new_sub_path = base_sub_path

        import_config_new = import_config.copy(update={"sub_path": new_sub_path})

        result_bundle = KiaraFileBundle.from_archive_file(
            archive_file=archive_zip, import_config=import_config_new
        )

        return result_bundle



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING, Any, Dict, Union

from kiara.models.values.value import ValueMap
from kiara_plugin.onboarding.modules import OnboardFileBundleModule, OnboardFileModule

if TYPE_CHECKING:
    from kiara.models.filesystem import FolderImportConfig, KiaraFile, KiaraFileBundle


class DownloadGithubFileModule(OnboardFileModule):
    """Download a single file from a github repo."""

    _module_type_name = "download.file.from.github"

    def create_onboard_inputs_schema(self) -> Dict[str, Any]:

        result: Dict[str, Dict[str, Any]] = {
            "user": {"type": "string", "doc": "The username/org-name."},
            "repo": {"type": "string", "doc": "The repository name."},
            "branch": {
                "type": "string",
                "doc": "The branch (or tag) name. If not specified, the 'main' branch name will be used.",
                "default": "main",
            },
            "path": {
                "type": "string",
                "doc": "The path to the file in the repository. Make sure not to specify a directory here, only a file path.",
            },
        }
        return result

    def retrieve_file(
        self, inputs: ValueMap, file_name: Union[str, None], attach_metadata: bool
    ) -> Any:

        from kiara_plugin.onboarding.utils.download import download_file

        user = inputs.get_value_data("user")
        repo = inputs.get_value_data("repo")
        branch = inputs.get_value_data("branch")
        sub_path = inputs.get_value_data("path")

        url = f"https://raw.githubusercontent.com/{user}/{repo}/{branch}/{sub_path}"

        result_file: KiaraFile = download_file(  # type: ignore
            url=url, file_name=file_name, attach_metadata=attach_metadata
        )
        return result_file


class DownloadGithbFileBundleModule(OnboardFileBundleModule):
    """Download a file bundle from a remote github repository.

    If 'sub_path' is not specified, the whole repo will be used.

    """

    _module_type_name = "download.file_bundle.from.github"

    def create_onboard_inputs_schema(self) -> Dict[str, Any]:
        result: Dict[str, Dict[str, Any]] = {
            "user": {"type": "string", "doc": "The username/org-name."},
            "repo": {"type": "string", "doc": "The repository name."},
            "branch": {
                "type": "string",
                "doc": "The branch (or tag) name. If not specified, the default branch will be used.",
                "optional": True,
            },
        }
        return result

    def retrieve_archive(
        self,
        inputs: ValueMap,
        bundle_name: Union[str, None],
        attach_metadata_to_bundle: bool,
        attach_metadata_to_files: bool,
        import_config: "FolderImportConfig",
    ) -> Union["KiaraFile", "KiaraFileBundle"]:

        from kiara_plugin.onboarding.utils.download import download_file

        user = inputs.get_value_data("user")
        repo = inputs.get_value_data("repo")
        branch = inputs.get_value_data("branch")
        if not branch:
            branch = "main"

        url = f"https://github.com/{user}/{repo}/archive/refs/heads/{branch}.zip"

        file_name = f"{repo}-{branch}.zip"
        result_file: KiaraFile = download_file(  # type: ignore
            url=url, file_name=file_name, attach_metadata=True, return_md5_hash=False
        )
        return result_file
